from ._stats import (SongbirdStats, SongbirdStatsDirFmt, SongbirdStatsFormat)
from ._method import multinomial
from ._summary import summarize_single, summarize_paired


__all__ = ['multinomial', 'summarize_single', 'summarize_paired',
           'SongbirdStats', 'SongbirdStatsFormat',
           'SongbirdStatsDirFmt']



from qiime2.plugin import SemanticType, model
from q2_types.sample_data import SampleData


# songbird stats summarizing loss and cv error
SongbirdStats = SemanticType('SongbirdStats',
                             variant_of=SampleData.field['type'])


class SongbirdStatsFormat(model.TextFileFormat):
    def validate(*args):
        pass


SongbirdStatsDirFmt = model.SingleFileDirectoryFormat(
    'SongbirdStatsDirFmt', 'stats.tsv', SongbirdStatsFormat)



# Parameter Descriptions
DESCS = {
    "table": "Input table of counts.",
    "metadata": "Sample metadata file with covariates of interest.",
    "formula": (
        'The statistical formula specifying covariates to be included in the '
        'model and their interactions.'
    ),
    "training-column": (
        'The column in the metadata file used to '
        'specify training and testing. These columns '
        'should be specifically labeled (Train) and (Test).'
    ),
    "num-random-test-examples": (
        'Number of random samples to hold out for cross-validation if '
        'a training column is not specified.'
    ),
    "epochs": 'The total number of iterations over the entire dataset.',
    "batch-size": (
        "The number of samples to be evaluated per training iteration."
    ),
    "differential-prior": (
        'Width of normal prior for the `differentials`, or '
        'the coefficients of the multinomial regression. '
        'Smaller values will force the coefficients towards zero. '
        'Values must be greater than 0.'
    ),
    'learning-rate': 'Gradient descent decay rate.',
    'clipnorm': "Gradient clipping size.",
    'min-sample-count': (
        "The minimum number of counts a sample needs for it to be included in "
        "the analysis."
    ),
    'min-feature-count': (
        "The minimum number of samples a feature needs to be observed in "
        "for it to be included in the analysis."
    ),
    "summary-interval": "Number of seconds before storing a summary.",
    "silent": (
        'Flag denoting whether to hide progress bar and TensorFlow warnings. '
        'If False, progress bar and warnings are displayed. If True, neither '
        'progress bar nor warnings are displayed.'
    ),
    # FYI: The following parameters are exclusive to the non-Q2 Songbird script
    "checkpoint-interval": 'Number of seconds before a saving a checkpoint.',
    "summary-dir": (
        'Summary directory to save regression results to. '
        'This will include a table of differentials under '
        '`differentials.tsv` that can be ranked, in addition '
        'to summaries that can be loaded into Tensorboard and '
        'checkpoints for recovering parameters during runtime.'
    ),
    "random-seed": (
        'The number to used to receive consistent results for the random  '
        'processes in the fitting procedure.'
    ),
}

DEFAULTS = {
    "training-column": None,
    "num-random-test-examples": 5,
    "epochs": 1000,
    "batch-size": 5,
    "differential-prior": 1.,
    'learning-rate': 1e-3,
    'clipnorm': 10.,
    'min-sample-count': 1000,
    'min-feature-count': 10,
    "checkpoint-interval": 3600,
    "summary-interval": 10,
    "summary-dir": "summarydir",
    "random-seed": 0,
    "silent": False,
}



import qiime2
from songbird.q2 import SongbirdStatsFormat
from songbird.q2.plugin_setup import plugin


@plugin.register_transformer
def _1(ff: SongbirdStatsFormat) -> qiime2.Metadata:
    return qiime2.Metadata.load(str(ff))


@plugin.register_transformer
def _2(obj: qiime2.Metadata) -> SongbirdStatsFormat:
    ff = SongbirdStatsFormat()
    obj.save(str(ff))
    return ff



import os
import tensorflow as tf
import numpy as np
import pandas as pd
from sklearn.utils import check_random_state
from skbio.stats.composition import clr_inv as softmax
from biom import Table
from patsy import dmatrix


def random_multinomial_model(num_samples, num_features,
                             reps=1,
                             low=2, high=10,
                             beta_mean=0,
                             beta_scale=5,
                             mu=1,
                             sigma=1,
                             seed=0):
    """ Generates a table using a random poisson regression model.

    Here we will be simulating microbial counts given the model, and the
    corresponding model priors.

    Parameters
    ----------
    num_samples : int
        Number of samples
    num_features : int
        Number of features
    tree : np.array
        Tree specifying orthonormal contrast matrix.
    low : float
        Smallest gradient value.
    high : float
        Largest gradient value.
    beta_mean : float
        Mean of beta prior (for regression coefficients)
    beta_scale : float
        Scale of beta prior (for regression coefficients)
    mu : float
        Mean sequencing depth (in log units)
    sigma : float
        Variance for sequencing depth

    Returns
    -------
    table : biom.Table
        Biom representation of the count table.
    metadata : pd.DataFrame
        DataFrame containing relevant metadata.
    beta : np.array
        Regression parameter estimates.
    """
    N = num_samples

    # generate all of the coefficient using the random poisson model
    state = check_random_state(seed)
    beta = state.normal(beta_mean, beta_scale, size=(2, num_features-1))

    X = np.hstack([np.linspace(low, high, num_samples // reps)]
                  for _ in range(reps))
    X = np.vstack((np.ones(N), X)).T
    phi = np.hstack((np.zeros((N, 1)), X @ beta))
    probs = softmax(phi)
    n = [mu] * N

    table = np.vstack(
        state.multinomial(n[i], probs[i, :])
        for i in range(N)
    ).T

    samp_ids = pd.Index(['S%d' % i for i in range(num_samples)],
                        name='sampleid')
    feat_ids = ['F%d' % i for i in range(num_features)]
    balance_ids = ['L%d' % i for i in range(num_features-1)]

    table = Table(table, feat_ids, samp_ids)
    metadata = pd.DataFrame(X, columns=['Ones', 'X'], index=samp_ids)
    beta = pd.DataFrame(beta.T, columns=['Intercept', 'beta'],
                        index=balance_ids)

    return table, metadata, beta


def _type_cast_to_float(df):
    """ Attempt to cast all of the values in dataframe to float.

    This will try to type cast all of the series within the
    dataframe into floats.  If a column cannot be type casted,
    it will be kept as is.

    Parameters
    ----------
    df : pd.DataFrame

    Returns
    -------
    pd.DataFrame
    """
    # TODO: Will need to improve this, as this is a very hacky solution.
    for c in df.columns:
        s = df[c]
        try:
            df[c] = s.astype(np.float64)
        except Exception:
            continue
    return df


def read_metadata(filepath):
    """ Reads in a sample metadata file

    Parameters
    ----------
    filepath: str
       The file path location of the sample metadata file

    Returns
    -------
    pd.DataFrame :
       The metadata table with inferred types.
    """
    metadata = pd.read_table(
        filepath, dtype=object)
    cols = metadata.columns
    metadata = metadata.set_index(cols[0])
    metadata = _type_cast_to_float(metadata.copy())

    return metadata


def match_and_filter(table, metadata, formula,
                     min_sample_count, min_feature_count):
    """ Matches and aligns biom and metadata tables.

    This will also return the patsy representation.

    Parameters
    ----------
    table : biom.Table
        Table of abundances
    metadata : pd.DataFrame
        Sample metadata

    Returns
    -------
    table : biom.Table
        Filtered biom table
    metadata : pd.DataFrame
        Sample metadata
    """
    # match them

    def sample_filter(val, id_, md):
        return id_ in metadata.index and np.sum(val) > min_sample_count

    def read_filter(val, id_, md):
        return np.sum(val > 0) > min_feature_count

    table = table.filter(sample_filter, axis='sample', inplace=False)
    table = table.filter(read_filter, axis='observation', inplace=False)

    metadata = metadata.loc[table.ids(axis='sample')]
    metadata = metadata.loc[~metadata.index.duplicated(keep='first')]

    def sort_f(xs):
        return [xs[metadata.index.get_loc(x)] for x in xs]

    table = table.sort(sort_f=sort_f, axis='sample')
    design = dmatrix(formula, metadata, return_type='dataframe')
    design = design.dropna()

    def design_filter(val, id_, md):
        return id_ in design.index

    table = table.filter(design_filter, axis='sample')
    return table, metadata, design


def split_training(dense_table, metadata, design, training_column=None,
                   num_random_test_examples=10, seed=None):

    if training_column is None:
        np.random.seed(seed)
        idx = np.random.random(design.shape[0])
        i = np.argsort(idx)[num_random_test_examples]

        threshold = idx[i]
        train_idx = ~(idx < threshold)
    else:
        train_idx = metadata.loc[design.index, training_column] == "Train"

    trainX = design.loc[train_idx].values
    testX = design.loc[~train_idx].values

    trainY = dense_table.loc[train_idx].values
    testY = dense_table.loc[~train_idx].values

    return trainX, testX, trainY, testY


def silence_output():
    # suppress profiling messages & compilation warnings
    # taken from:
    # https://stackoverflow.com/questions/47068709/your-cpu-supports-
    #     instructions-that-this-tensorflow-binary-was-not-compiled-to-u
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

    # suppress deprecation warnings
    # taken from https://github.com/tensorflow/tensorflow/issues/27023
    tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)



import skbio
import qiime2
import pandas as pd
import numpy as np
import biom
import tensorflow as tf
from skbio import OrdinationResults
from songbird.multinomial import MultRegression
from songbird.util import match_and_filter, split_training, silence_output
from songbird.parameter_info import DEFAULTS
from qiime2.plugin import Metadata


def multinomial(table: biom.Table,
                metadata: Metadata,
                formula: str,
                training_column: str = DEFAULTS["training-column"],
                num_random_test_examples: int = (
                    DEFAULTS["num-random-test-examples"]
                ),
                epochs: int = DEFAULTS["epochs"],
                batch_size: int = DEFAULTS["batch-size"],
                differential_prior: float = DEFAULTS["differential-prior"],
                learning_rate: float = DEFAULTS["learning-rate"],
                clipnorm: float = DEFAULTS["clipnorm"],
                min_sample_count: int = DEFAULTS["min-sample-count"],
                min_feature_count: int = DEFAULTS["min-feature-count"],
                summary_interval: int = DEFAULTS["summary-interval"],
                random_seed: int = DEFAULTS["random-seed"],
                silent: bool = DEFAULTS["silent"],
                ) -> (
                    pd.DataFrame, qiime2.Metadata, skbio.OrdinationResults
                ):

    if silent:
        silence_output()

    # load metadata and tables
    metadata = metadata.to_dataframe()
    # match them
    table, metadata, design = match_and_filter(
        table, metadata,
        formula, min_sample_count, min_feature_count
    )

    # convert to dense representation
    dense_table = table.to_dataframe().to_dense().T

    # split up training and testing
    trainX, testX, trainY, testY = split_training(
        dense_table, metadata, design,
        training_column, num_random_test_examples,
        seed=random_seed,
    )

    model = MultRegression(learning_rate=learning_rate, clipnorm=clipnorm,
                           beta_mean=differential_prior,
                           batch_size=batch_size,
                           save_path=None)

    with tf.Graph().as_default(), tf.Session() as session:
        tf.set_random_seed(random_seed)
        model(session, trainX, trainY, testX, testY)

        loss, cv, its = model.fit(
            epochs=epochs,
            summary_interval=summary_interval,
            checkpoint_interval=None,
            silent=silent)

    md_ids = np.array(design.columns)
    obs_ids = table.ids(axis='observation')

    beta_ = np.hstack((np.zeros((model.p, 1)), model.B))
    beta_ = beta_ - beta_.mean(axis=1).reshape(-1, 1)

    differentials = pd.DataFrame(
        beta_.T, columns=md_ids, index=obs_ids,
    )
    differentials.index.name = 'featureid'

    convergence_stats = pd.DataFrame(
        {
            'loss': loss,
            'cross-validation': cv,
            'iteration': its
        }
    )

    convergence_stats.index.name = 'id'
    convergence_stats.index = convergence_stats.index.astype(np.str)

    c = convergence_stats['loss'].astype(np.float)
    convergence_stats['loss'] = c

    c = convergence_stats['cross-validation'].astype(np.float)
    convergence_stats['cross-validation'] = c

    c = convergence_stats['iteration'].astype(np.int)
    convergence_stats['iteration'] = c

    # regression biplot
    if differentials.shape[-1] > 1:
        u, s, v = np.linalg.svd(differentials)
        pc_ids = ['PC%d' % i for i in range(len(s))]
        samples = pd.DataFrame(u[:, :len(s)] @ np.diag(s),
                               columns=pc_ids, index=differentials.index)
        features = pd.DataFrame(v.T[:, :len(s)],
                                columns=pc_ids, index=differentials.columns)
        short_method_name = 'regression_biplot'
        long_method_name = 'Multinomial regression biplot'
        eigvals = pd.Series(s, index=pc_ids)
        proportion_explained = eigvals**2 / (eigvals**2).sum()
        biplot = OrdinationResults(
            short_method_name, long_method_name, eigvals,
            samples=samples, features=features,
            proportion_explained=proportion_explained)
    else:
        # this is to handle the edge case with only intercepts
        biplot = OrdinationResults('', '', pd.Series(), pd.DataFrame())

    return differentials, qiime2.Metadata(convergence_stats), biplot



import os
import tensorflow as tf
from tensorflow.contrib.distributions import Multinomial, Normal
from tqdm import tqdm
import time
import datetime
import numpy as np


class MultRegression(object):

    def __init__(self, beta_mean=0, beta_scale=1,
                 batch_size=5, learning_rate=0.001, beta_1=0.9, beta_2=0.99,
                 clipnorm=10., save_path=""):
        """ Build a tensorflow model

        Returns
        -------
        loss : tf.Tensor
           The log loss of the model.

        Notes
        -----
        If save_path is None, there won't be anything saved
        """
        if save_path == "":
            basename = "logdir"
            suffix = datetime.datetime.now().strftime("%y%m%d_%H%M%S")
            save_path = "_".join([basename, suffix])

        self.learning_rate = learning_rate
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.batch_size = batch_size
        self.clipnorm = clipnorm

        self.beta_mean = beta_mean
        self.beta_scale = beta_scale
        self.save_path = save_path

    def __call__(self, session, trainX, trainY, testX, testY):
        """ Initialize the actual graph

        Parameters
        ----------
        session : tf.Session
            Tensorflow session
        trainX : np.array
            Input training design matrix.
        trainY : np.array
            Output training OTU table, where rows are samples and columns are
            observations.
        testX : np.array
            Input testing design matrix.
        testY : np.array
            Output testing OTU table, where rows are samples and columns are
            observations.
        """
        self.session = session
        self.N, self.p = trainX.shape
        self.D = trainY.shape[1]
        holdout_size = testX.shape[0]

        # Place holder variables to accept input data
        self.X_ph = tf.constant(trainX, dtype=tf.float32, name='G_ph')
        self.Y_ph = tf.constant(trainY, dtype=tf.float32, name='Y_ph')
        self.X_holdout = tf.constant(testX, dtype=tf.float32, name='G_holdout')
        self.Y_holdout = tf.constant(testY, dtype=tf.float32, name='Y_holdout')

        batch_ids = tf.multinomial(tf.ones([1, self.N]), self.batch_size)
        sample_ids = tf.squeeze(batch_ids)

        Y_batch = tf.gather(self.Y_ph, sample_ids, axis=0)
        X_batch = tf.gather(self.X_ph, sample_ids, axis=0)

        total_count = tf.reduce_sum(Y_batch, axis=1)
        holdout_count = tf.reduce_sum(self.Y_holdout, axis=1)

        # Define PointMass Variables first
        self.qbeta = tf.Variable(
            tf.random_normal([self.p, self.D-1]), name='qB')

        # regression coefficents distribution
        beta = Normal(loc=tf.zeros([self.p, self.D-1]) + self.beta_mean,
                      scale=tf.ones([self.p, self.D-1]) * self.beta_scale,
                      name='B')

        eta = tf.matmul(X_batch, self.qbeta, name='eta')

        phi = tf.nn.log_softmax(
            tf.concat(
                [tf.zeros([self.batch_size, 1]), eta], axis=1), name='phi'
        )

        Y = Multinomial(total_count=total_count, logits=phi, name='Y')

        # cross validation
        with tf.name_scope('accuracy'):
            pred = tf.reshape(
                holdout_count, [-1, 1]) * tf.nn.softmax(
                    tf.concat([
                        tf.zeros([holdout_size, 1]),
                        tf.matmul(self.X_holdout, self.qbeta)
                    ], axis=1), name='phi'
                )

            self.cv = tf.reduce_mean(
                tf.squeeze(tf.abs(pred - self.Y_holdout))
            )
            tf.summary.scalar('cv_error', self.cv)

        self.loss = -(tf.reduce_sum(beta.log_prob(self.qbeta)) +
                      tf.reduce_sum(Y.log_prob(Y_batch)) *
                      (self.N / self.batch_size))

        optimizer = tf.train.AdamOptimizer(
            self.learning_rate, beta1=self.beta_1, beta2=self.beta_2)

        gradients, variables = zip(*optimizer.compute_gradients(self.loss))
        self.gradients, _ = tf.clip_by_global_norm(gradients, self.clipnorm)
        self.train = optimizer.apply_gradients(zip(gradients, variables))

        tf.summary.scalar('loss', self.loss)
        tf.summary.histogram('qbeta', self.qbeta)
        self.merged = tf.summary.merge_all()
        if self.save_path is not None:
            self.writer = tf.summary.FileWriter(self.save_path,
                                                self.session.graph)
        else:
            self.writer = None
        tf.global_variables_initializer().run()

    def fit(self, epochs=10, summary_interval=100, checkpoint_interval=3600,
            silent=False):
        """ Fits the model.

        Parameters
        ----------
        epochs : int
           Number of epochs to train
        summary_interval : float
           Number of seconds until a summary is recorded
        checkpoint_interval : int
           Number of seconds until a checkpoint is recorded
        silent : bool
           Flag denoting whether to suppress progress bar and TensorFlow
           warnings. If True, show neither progress bar nor warnings. If False
           (default), show both.

        Returns
        -------
        loss: np.array
            log likelihood loss.
        cv : np.array
            cross validation loss
        iter_n : np.array
            iterations
        """
        num_iter = (self.N // self.batch_size) * epochs
        cv = None
        last_checkpoint_time = 0
        last_summary_time = 0
        saver = tf.train.Saver()
        loss = []
        cv = []
        iter_n = []

        if silent:
            iter_range = range(0, num_iter)
        else:
            iter_range = tqdm(range(0, num_iter))

        for i in iter_range:
            now = time.time()

            if now - last_summary_time > summary_interval:
                run_options = tf.RunOptions(
                    trace_level=tf.RunOptions.FULL_TRACE)
                run_metadata = tf.RunMetadata()
                _, summary, train_loss, test_cv, grads = self.session.run(
                    [self.train, self.merged,
                     self.loss, self.cv, self.gradients],
                    options=run_options,
                    run_metadata=run_metadata
                )
                cv.append(test_cv)
                loss.append(train_loss)
                iter_n.append(i)

                if self.writer is not None:
                    self.writer.add_run_metadata(run_metadata, 'step%d' % i)
                    self.writer.add_summary(summary, i)
                last_summary_time = now
            else:
                _, summary, train_loss, grads = self.session.run(
                    [self.train, self.merged, self.loss, self.gradients],
                )
                if self.writer is not None:
                    self.writer.add_summary(summary, i)

            if checkpoint_interval is not None:
                if now - last_checkpoint_time > checkpoint_interval:
                    saver.save(self.session,
                               os.path.join(self.save_path, "model.ckpt"),
                               global_step=i)
                last_checkpoint_time = now

        train_loss, test_cv, B = self.session.run(
            [self.loss, self.cv, self.qbeta]
        )
        cv.append(test_cv)
        loss.append(train_loss)
        iter_n.append(i)

        self.B = B
        return np.array(loss), np.array(cv), np.array(iter_n)



# ----------------------------------------------------------------------------
# Copyright (c) 2016--, gneiss development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------
import importlib
import qiime2.plugin
import qiime2.sdk
from songbird import __version__

from qiime2.plugin import (Str, Properties, Int, Float,  Metadata, Bool)
from q2_types.feature_table import FeatureTable, Frequency
from q2_types.ordination import PCoAResults
from q2_types.sample_data import SampleData
from q2_types.feature_data import (FeatureData, Differential)
from songbird.q2 import (
    SongbirdStats, SongbirdStatsFormat, SongbirdStatsDirFmt,
    multinomial, summarize_single, summarize_paired
)
from songbird.parameter_info import DESCS

citations = qiime2.plugin.Citations.load('citations.bib', package='songbird')

plugin = qiime2.plugin.Plugin(
    name='songbird',
    version=__version__,
    website="https://github.com/mortonjt/songbird",
    citations=[citations['MortonMarotz2019']],
    short_description=('Plugin for differential abundance analysis '
                       'via count-based models.'),
    description=('This is a QIIME 2 plugin supporting statistical models on '
                 'feature tables and metadata.'),
    package='songbird')

plugin.methods.register_function(
    function=multinomial,
    inputs={'table': FeatureTable[Frequency]},
    parameters={
        'metadata': Metadata,
        'formula': Str,
        'training_column': Str,
        'num_random_test_examples': Int,
        'epochs': Int,
        'batch_size': Int,
        'differential_prior': Float,
        'learning_rate': Float,
        'clipnorm': Float,
        'min_sample_count': Int,
        'min_feature_count': Int,
        'summary_interval': Float,
        'random_seed': Int,
        'silent': Bool,
    },
    outputs=[
        ('differentials', FeatureData[Differential]),
        ('regression_stats', SampleData[SongbirdStats]),
        ('regression_biplot', PCoAResults % Properties('biplot'))
    ],
    input_descriptions={
        'table': DESCS["table"],
    },
    output_descriptions={
        'differentials': ('Output differentials learned from the '
                          'multinomial regression.'),
        'regression_stats': ('Summary information about the loss '
                             'and cross validation error over iterations.'),
        'regression_biplot': ('A biplot of the regression coefficients')
    },
    parameter_descriptions={
        'metadata': DESCS["metadata"],
        'formula': DESCS["formula"],
        "training_column": DESCS["training-column"],
        'num_random_test_examples': DESCS["num-random-test-examples"],
        'epochs': DESCS["epochs"],
        'batch_size': DESCS["batch-size"],
        'differential_prior': DESCS["differential-prior"],
        'learning_rate': DESCS["learning-rate"],
        "clipnorm": DESCS["clipnorm"],
        "min_sample_count": DESCS["min-sample-count"],
        "min_feature_count": DESCS["min-feature-count"],
        "summary_interval": DESCS["summary-interval"],
        "random_seed": DESCS["random-seed"],
        "silent": (DESCS["silent"] + " (Only has an impact when using this "
                   "command with the --verbose option or through the Qiime2 "
                   "Artifact API)"),
    },
    name='Multinomial regression',
    description=("Performs multinomial regression and calculates "
                 "rank differentials for organisms with respect to the "
                 "covariates of interest."),
    citations=[]
)

plugin.visualizers.register_function(
    function=summarize_single,
    inputs={
        'regression_stats': SampleData[SongbirdStats]
    },
    parameters={},
    input_descriptions={
        'regression_stats': (
            "Summary information produced by running "
            "`qiime songbird multinomial`."
        )
    },
    parameter_descriptions={
    },
    name='Regression summary statistics',
    description=(
        "Visualize the convergence statistics from running multinomial "
        "regression, giving insight into how the model fit to your data."
    )
)

plugin.visualizers.register_function(
    function=summarize_paired,
    inputs={
        'regression_stats': SampleData[SongbirdStats],
        'baseline_stats': SampleData[SongbirdStats]
    },
    parameters={},
    input_descriptions={

        'regression_stats': (
            "Summary information for the reference model, produced by running "
            "`qiime songbird multinomial`."
        ),
        'baseline_stats': (
            "Summary information for the baseline model, produced by running "
            "`qiime songbird multinomial`."
        )

    },
    parameter_descriptions={
    },
    name='Paired regression summary statistics',
    description=(
        "Visualize the convergence statistics from two runs of multinomial "
        "regression, giving insight into how the models fit to your data. "
        "The produced visualization includes a 'pseudo-Q-squared' value."
    )
)

# Register types
plugin.register_formats(SongbirdStatsFormat, SongbirdStatsDirFmt)
plugin.register_semantic_types(SongbirdStats)
plugin.register_semantic_type_to_format(
    SampleData[SongbirdStats], SongbirdStatsDirFmt)

importlib.import_module('songbird.q2._transformer')



import os
import qiime2
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


def _convergence_plot(regression, baseline, ax0, ax1):
    iterations = np.array(regression['iteration'])

    ax0.plot(iterations[1:],
             np.array(regression['cross-validation'].values)[1:],
             label='model')
    ax0.set_ylabel('Cross validation score', fontsize=14)
    ax0.set_xlabel('# Iterations', fontsize=14)

    ax1.plot(iterations[1:],
             np.array(regression['loss'])[1:], label='model')
    ax1.set_ylabel('Loss', fontsize=14)
    ax1.set_xlabel('# Iterations', fontsize=14)

    if baseline is not None:
        iterations = baseline['iteration']

        ax0.plot(iterations[1:],
                 np.array(baseline['cross-validation'].values)[1:],
                 label='baseline')
        ax0.set_ylabel('Cross validation score', fontsize=14)
        ax0.set_xlabel('# Iterations', fontsize=14)
        ax0.legend()

        ax1.plot(iterations[1:],
                 np.array(baseline['loss'])[1:], label='baseline')
        ax1.set_ylabel('Loss', fontsize=14)
        ax1.set_xlabel('# Iterations', fontsize=14)
        ax1.legend()


def _summarize(output_dir: str, regression: pd.DataFrame,
               baseline: pd.DataFrame = None):

    """ Helper method for generating summary pages

    Parameters
    ----------
    output_dir : str
       Name of output directory
    regression : pd.DataFrame
       Regression summary with column names
       ['loss', 'cross-validation']
    baseline : pd.DataFrame
       Baseline regression summary with column names
       ['loss', 'cross-validation']. Defaults to None (i.e. if only a single
       set of regression stats will be summarized).

    Note
    ----
    There may be synchronizing issues if different summary intervals
    were used between analyses. For predictable results, try to use the
    same summary interval.
    """
    fig, ax = plt.subplots(2, 1, figsize=(10, 10))
    if baseline is None:
        _convergence_plot(regression, None, ax[0], ax[1])
        q2 = None
    else:

        _convergence_plot(regression, baseline, ax[0], ax[1])

        # this provides a pseudo-r2 commonly provided in the context
        # of logistic / multinomail regression (proposed by Cox & Snell)
        # http://www3.stat.sinica.edu.tw/statistica/oldpdf/a16n39.pdf
        end = min(10, len(regression.index))
        # trim only the last 10 numbers

        # compute a q2 score, which is commonly used in
        # partial least squares for cross validation
        l0 = np.mean(baseline['cross-validation'][-end:])
        lm = np.mean(regression['cross-validation'][-end:])
        q2 = 1 - lm / l0

    plt.tight_layout()
    fig.savefig(os.path.join(output_dir, 'convergence-plot.svg'))
    fig.savefig(os.path.join(output_dir, 'convergence-plot.pdf'))

    index_fp = os.path.join(output_dir, 'index.html')
    with open(index_fp, 'w') as index_f:
        index_f.write('<html><body>\n')
        index_f.write('<h1>Convergence summary</h1>\n')
        index_f.write(
            "<p>If you don't see anything in these plots, you probably need "
            "to decrease your <kbd>--p-summary-interval</kbd>. Try setting "
            "<kbd>--p-summary-interval 1</kbd>, which will record the loss at "
            "every second.</p>\n"
        )
        index_f.write(
            "<p>For information about <strong>how to interpret these "
            "plots</strong>, please see "
            '<a href="https://github.com/biocore/songbird#interpreting-model-fitting">this section</a> '  # noqa
            "of the Songbird README.</p>"
        )
        index_f.write(
            "<p>For information about <strong>how to adjust Songbird's "
            "parameters to get the model to fit reasonably well to your "
            "dataset</strong>, please see "
            '<a href="https://github.com/biocore/songbird#adjusting-parameters">this section</a> '  # noqa
            "of the Songbird README.</p>"
        )
        if q2 is not None:
            index_f.write(
                '<p><strong>'
                '<a href="https://github.com/biocore/songbird#explaining-q2">'
                'Pseudo Q-squared:</a></strong> %f</p>\n' % q2
            )

        index_f.write(
            '<img src="convergence-plot.svg" alt="convergence_plots">'
        )
        index_f.write('<a href="convergence-plot.pdf">')
        index_f.write('Download as PDF</a><br>\n')


def summarize_single(output_dir: str, regression_stats: qiime2.Metadata):
    _summarize(output_dir, regression_stats.to_dataframe())


def summarize_paired(output_dir: str,
                     regression_stats: qiime2.Metadata,
                     baseline_stats: qiime2.Metadata):
    _summarize(output_dir,
               regression_stats.to_dataframe(),
               baseline_stats.to_dataframe())
